/*
 * Copyright (C) 2011-2012 Koji Suzuki
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 2.1 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

//
// Module Name: alu8
//   ver 001
// 
// ALU based 74181
//
// (1) normalized C_IN , C_OUT
// (2) enhanced Function :
//
//     S[7:4]   S[3:0]
//
//     1001      FS[3:0]  -- normal 181
//     0110      FS[3:0]  -- invert A
//     FS[3:0]   1001     -- swap A,B
//     FS[3:0]   0110     -- invert B and swap A,B
//
//
//   FS[3:0]      M = 1      C_IN = 0    C_IN = 1
//                          (no carry)  (with carry)
//   0 0 0 0       ~A            A          A + 1
//   0 0 0 1    ~(A | B)  
//   0 0 1 0     ~A & B  
//   0 0 1 1        0           -1            0
//   0 1 0 0    ~(A & B)   
//   0 1 0 1       ~B    
//   0 1 1 0      A ^ B        A - B - 1    A - B   
//   0 1 1 1      A & ~B       
//   1 0 0 0     ~A | B
//   1 0 0 1    ~(A ^ B)       A + B        A + B + 1
//   1 0 1 0        B           
//   1 0 1 1      A & B
//   1 1 0 0        1          A + A        A + A + 1
//   1 1 0 1      A | ~B
//   1 1 1 0      A | B
//   1 1 1 1        A          A - 1          A
//
`define ALU8_USE_ALU4
module alu8 (
		 input   C_IN      // Carry Input
		, input  [7:0] A   // DATA IN
		, input  [7:0] B   // DATA IN
		, output [7:0] F   // DATA OUT
		, output  OVR      // Overflow
		, output  EQ       // F != 0
		, output  C_HALF   // carry-look-ahead
		, output  C_OUT    // carry-look-ahead

		, output  P        // carry-propagate (INV)
		, output  G        // carry-generate (INV)

		, input   M        // Mode Control
		, input   [7:0] S  // Function Select
		);
	wire C_HALF;
`ifdef ALU8_USE_ALU4
	wire EQ_0, EQ_1;
	alu4 alu_lo (.C_IN(C_IN), .A(A[3:0]), .B(B[3:0]), .F(F[3:0])
		     , .M(M), .S(S[7:0]), .C_OUT(C_HALF), .EQ(EQ_0)
		     , .P(PL), .G(GL) );
	alu4 alu_hi (.C_IN(C_HALF), .A(A[7:4]), .B(B[7:4]), .F(F[7:4])
		     , .M(M), .S(S[7:0]), .C_OUT(C_OUT), .EQ(EQ_1), .OVR(OVR)
		     , .P(PH), .G(GH) );
	assign EQ = EQ_0 & EQ_1;
	assign P = PL | PH;
	assign G = GH & (PH & GL );
`else
    wire [7:0] D;
    wire [7:0] C;
    assign D[0] = ~( A[0] & S[4] | ~A[0] & S[5] | B[0] & S[0] | ~B[0] & S[1] );
    assign D[1] = ~( A[1] & S[4] | ~A[1] & S[5] | B[1] & S[0] | ~B[1] & S[1] );
    assign D[2] = ~( A[2] & S[4] | ~A[2] & S[5] | B[2] & S[0] | ~B[2] & S[1] );
    assign D[3] = ~( A[3] & S[4] | ~A[3] & S[5] | B[3] & S[0] | ~B[3] & S[1] );
    assign D[4] = ~( A[4] & S[4] | ~A[3] & S[5] | B[4] & S[0] | ~B[4] & S[1] );
    assign D[5] = ~( A[5] & S[4] | ~A[3] & S[5] | B[5] & S[0] | ~B[5] & S[1] );
    assign D[6] = ~( A[6] & S[4] | ~A[3] & S[5] | B[6] & S[0] | ~B[6] & S[1] );
    assign D[7] = ~( A[7] & S[4] | ~A[3] & S[5] | B[7] & S[0] | ~B[7] & S[1] );

    assign C[0] = ~((~A[0] & S[6] | A[0] & S[7])&(~B[0] & S[2] | B[0] & S[3]));
    assign C[1] = ~((~A[1] & S[6] | A[1] & S[7])&(~B[1] & S[2] | B[1] & S[3]));
    assign C[2] = ~((~A[2] & S[6] | A[2] & S[7])&(~B[2] & S[2] | B[2] & S[3]));
    assign C[3] = ~((~A[3] & S[6] | A[3] & S[7])&(~B[3] & S[2] | B[3] & S[3]));
    assign C[4] = ~((~A[4] & S[6] | A[4] & S[7])&(~B[4] & S[2] | B[4] & S[3]));
    assign C[5] = ~((~A[5] & S[6] | A[5] & S[7])&(~B[5] & S[2] | B[5] & S[3]));
    assign C[6] = ~((~A[6] & S[6] | A[6] & S[7])&(~B[6] & S[2] | B[6] & S[3]));
    assign C[7] = ~((~A[7] & S[6] | A[7] & S[7])&(~B[7] & S[2] | B[7] & S[3]));

    wire C_OUT1 = ~( D[0] | ~C_IN & C[0]);
    wire C_OUT2 = ~( D[1] | C[1] & D[0] | ~C_IN & C[0] & C[1]);
    wire C_OUT3 = ~( D[2] | C[2] & ( D[1] | C[1] & D[0])
		     | ~C_IN & C[0] & C[1] & C[2]);

    wire PL = ~( C[3] & C[2] & C[1] & C[0] );
    wire GL = ~(D[3] | C[3] & ( D[2] | C[2] & ( D[1] | C[1] & D[0])));
    assign C_HALF = ~(~GL | ~C_IN & C[0] & C[1] & C[2] & C[3]);

    assign F[0] = ~( ~M & ~C_IN ) ^ D[0] ^ C[0];
    assign F[1] = ~( ~M & ~C_OUT1 ) ^ D[1] ^ C[1];
    assign F[2] = ~( ~M & ~C_OUT2 ) ^ D[2] ^ C[2];
    assign F[3] = ~( ~M & ~C_OUT3 ) ^ D[3] ^ C[3];

    wire C_OUT5 = ~( D[4] | ~C_HALF & C[4]);
    wire C_OUT6 = ~( D[5] | C[5] & D[4] | ~C_HALF & C[4] & C[5]);
    wire C_OUT7 = ~( D[6] | C[6] & ( D[5] | C[5] & D[4])
		     | ~C_HALF & C[4] & C[5] & C[6]);

    wire PH = ~( C[7] & C[6] & C[5] & C[4] );
    wire GH = ~(D[7] | C[7] & ( D[6] | C[6] & ( D[5] | C[5] & D[4])));
    assign C_OUT = ~(~GH | ~C_HALF & C[4] & C[5] & C[6] & C[7]);

    assign F[4] = ~( ~M & ~C_HALF ) ^ D[4] ^ C[4];
    assign F[5] = ~( ~M & ~C_OUT5 ) ^ D[5] ^ C[5];
    assign F[6] = ~( ~M & ~C_OUT6 ) ^ D[6] ^ C[6];
    assign F[7] = ~( ~M & ~C_OUT7 ) ^ D[7] ^ C[7];

    assign P = PL | PH;
    assign G = GH & (PH & GL );

    assign OVR = ~M & (C_OUT7 ^ C_OUT);
    assign EQ = F[0] & F[1] & F[2] & F[3] & F[4] & F[5] & F[6] & F[7];
`endif
endmodule
