/*
 * Copyright (C) 2011 Koji Suzuki
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 2.1 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
// Original: 
// Minimal 8 Bit CPU (VHDL version)
//  rev 15102001
//  t.boescke@tuhh.de
//
// Instruction set
//   NOR : 00AAAAAA : Accu <= ~( Accu | mem[AAAAAA] )
//   ADD : 01AAAAAA : Accu <= Accu + mem[AAAAAA], update carry
//   STA : 10AAAAAA : mem[AAAAAA] <= Accu
//   JCC : 11DDDDDD : PC <= DDDDDD if (carry == 0) , carry <= 0
//
// Macro set
//   CLR      : NOR allone                 : 
//   LDA mem  : NOR allone, ADD mem        : Accu <= memory[mem]
//   NOT      : Nor zero                   : 
//   JMP dst  : JCC dst, JCC dst           : 
//   JCS dst  : JCC *+2 , JCC dst          : Jump if carry set
//   SUB mem  : NOR zero, ADD mem ADD one  : Accu -= memory[mem]
//       memory[allone] : 0xFF
//       memory[zero]   : 0x00
//       memory[one]    : 0x01
//
//  XO2-256    modified ROM/RAM    XO2-256
//  27     B   1            24  VCC  
//  28     G   2         x  23  DIG1    25
//  29         3         x  22  DIG2    23
//  30         4            21  DIG3    21
//  32         5            20  DIG4    20
//   1         6          x 19  CLK_1HZ 17
//   4     C   7            18  BTN_DN  16
//   5     D   8          x 17  N.C.
//   8     E   9            16  BTN_UP  14
//   9     A   10           15  BTN_SEL 13
//  10     F   11           14  CLK     12
//      GND  12             13  CLK_OUT 11

module mcpu2_efb (
		input CLK_3X
		,input RST // active high
                ,output WE
                ,inout SCL
                ,inout SDA
	);

    reg [7:0] I_DATA;
    wire [7:0] O_DATA;
    wire [7:0] mem_data;
    wire [11:0] code_data;

    wire [7:0] MEM_DATA;
    wire [7:0] WB_DATA;

    reg [8:0] akku = 0;
    reg [6:0] adreg = 0;
    reg [6:0] pc = 0;
    reg [2:0] states;

    assign O_DATA = akku[7:0];
    assign WE = (states == 1) & ~RST;

    wire WB_SEL = (adreg[6] == 1'b0) & (adreg[5:4] != 2'b11) ;
    wire MEM_SEL = ~WB_SEL;

    reg master_clk = 1'b0;
    wire CLK = master_clk;
    
    reg [1:0] wb_stat = 0;
    wire [7:0] wb_adr = { 2'b01, adreg[5:0] };
    wire [7:0] wb_data;

    wire wb_cyc = CLK & WB_SEL;
    wire wb_stb = CLK & WB_SEL;

    wire RST_NEG = ~RST;
    EFB_tc_i2c efb_impl (.wb_clk_i(CLK_3X), .wb_rst_i(RST)
    		  , .wb_adr_i(wb_adr[7:0]), .wb_dat_i(O_DATA[7:0]), .wb_dat_o(wb_data[7:0])
		  , .wb_we_i(WE), .wb_cyc_i(wb_cyc), .wb_stb_i(wb_stb)
		  , .tc_clki(CLK), .tc_rstn(RST_NEG)
		  , .i2c1_scl(SCL), .i2c1_sda(SDA)
		);

    wire WE_mem = WE & MEM_SEL;
    sram sram_impl (.CLK(CLK), .WEB(WE_mem), .ADDRB(adreg[5:0])
			, .DIB(O_DATA[7:0]) , .DOB(mem_data[7:0]) );

    rom rom_impl (.CLK(CLK), .ADDRA(adreg[6:0]), .DOA(code_data[11:0]) );

    always @(posedge CLK_3X)
    begin
	if (wb_stat == 0)
	  begin
		master_clk <= 1'b1;
		wb_stat <= 1;
	  end
	else if (wb_stat == 1)
	  begin
		master_clk <= 1'b1;
		wb_stat <= 2;
	  end
	else
	  begin
		master_clk <= 1'b0;
		wb_stat <= 0;
		I_DATA <= MEM_SEL ? mem_data : wb_data;
	  end
    end

    always @(RST, posedge CLK)
    begin
	if (~RST)
	    begin
		adreg <= 0;
		states <= 0;
		akku <= 0;
		pc <= 0;
	    end
	else
	    begin
		if (states == 0)
		    begin
			pc <= adreg + 1;
			adreg <= code_data[7:0];
		    end
		else    adreg <= pc;

		if (states == 2)
			akku <= { 1'b0 , akku[7:0] } + { 1'b0, I_DATA };
		else if (states == 3)
			akku[7:0] <= ~( akku[7:0] | I_DATA );
		else if  (states == 5)
			akku[8] <= 1'b0;

		if (states != 0)
			states <= 0;
		else if ((code_data[11:10] == 3) & akku[8])
			states <= 5;
		else
			states <= { 1'b0 , ~code_data[11:10] };
	    end
    end

endmodule

`define RAM_INCLUDE_DAT
// `define RAM_USE_CLK

module sram # (
//              parameter SIZE = 16
                parameter SIZE = 64
//              parameter SIZE = 128
//		parameter SIZE = 256
) (
                input CLK,
		input WEB,
                input [5:0] ADDRB,
                input [7:0] DIB,
                output [7:0] DOB
);
    reg  [7:0] mem [0:SIZE-1];
`ifdef RAM_USE_CLK
    reg [7:0] r_addr;
    always @(negedge CLK)
    begin
        if (!WEB) mem[ADDRB] <= DIB;
	r_addr <= ADDRB;
    end
    assign DOB = mem[r_addr];
`else
    always @(posedge WEB)
    begin
        mem[ADDRB] <= DIB;
    end
    assign DOB = mem[ADDRB];
`endif

`ifdef RAM_INCLUDE_DAT
  initial
    begin
       $readmemh("mcpu_ram.mem" , mem, 0 , SIZE-1);
    end
`endif

endmodule

`define ROM_USE_CLK
module rom # (
//              parameter SIZE = 16
//              parameter SIZE = 64
//              parameter SIZE = 128
		parameter SIZE = 256
) (
                input CLK,
                input [6:0] ADDRA,
                output [11:0] DOA
);
    reg  [11:0] mem [0:SIZE-1];
`ifdef ROM_USE_CLK
    reg [6:0] r_addr;
    always @(negedge CLK)
    begin
	r_addr <= ADDRA;
    end
    assign DOA = mem[r_addr];
`else
    assign DOA = mem[ADDRA];
`endif

  initial
    begin
       $readmemh("mcpu2_code64.mem" , mem, 0 , SIZE-1);
//       $readmemh("mcpu2_code128.mem" , mem, 0 , SIZE-1);
    end

endmodule
